package service

import (
	"context"
	"crypto/rand"
	"crypto/sha256"
	"encoding/hex"
	"encoding/json"
	"errors"
	"strings"
	"time"

	account "gitlab.mashroom.com/mashroom/backend/account/api"
	auth "gitlab.mashroom.com/mashroom/backend/auth/api"
	"gitlab.mashroom.com/mashroom/backend/auth/internal/model"
	"gitlab.mashroom.com/mashroom/backend/auth/internal/storage"
	"gitlab.mashroom.com/mashroom/backend/auth/keycloak"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	invitation "gitlab.mashroom.com/mashroom/backend/invitation/api"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
)

var (
	errResetTooSoon  = grpcerr.New(grpcerr.NoAuth, "reset reattempt too soon")
	errUserNotFound  = grpcerr.New(grpcerr.NotFound, "user not found")
	errInvalidToken  = grpcerr.New(grpcerr.InvalidData, "invalid token")
	errInactiveToken = grpcerr.New(grpcerr.NoAuth, "no active token")
)

// keycloakTokenIdentity for specified access token selects an unique value to identify user across Keycloak realm
// or returns an empty string. Checks token claims in the following order: email, username, preferred_username.
// NOTE: Keycloak's realm user ID value is contained in token.Sub claim and it is related as 1-1 to account.Account.ID
func keycloakTokenIdentity(introspect *storage.SessionIntrospection) string {
	switch {
	default:
		return ""
	case introspect == nil:
		return ""
	case introspect.Email != "":
		// email is primary user identity
		return introspect.Email
	case introspect.Username != "":
		// some old KK accounts have empty email value, but non-empty username which is expected equal the email
		return introspect.Username
	case introspect.PreferredUsername != "":
		// introspect retrospection may not return username directly so we can use mapped OpenID value for username
		return introspect.PreferredUsername
	}
}

func (srv *serverImpl) Auth(ctx context.Context, req *auth.AuthRequest) (*auth.AuthResponse, error) {
	token, err := srv.Storage.SessionIntrospect(ctx, req.GetAccessToken())
	if err != nil {
		log.WithError(err).Error()
		return nil, errInactiveToken
	}

	if !token.Active {
		return nil, errInactiveToken
	}

	acc, _, err := srv.accountLoadID(ctx, token.Sub, req.GetAdminSite())
	if err != nil {
		return nil, err
	}

	return &auth.AuthResponse{
		Account: acc,
		Session: token.SessionState,
		Roles:   token.RealmAccess.Roles,
	}, nil
}

func (srv *serverImpl) preLogin(
	ctx context.Context,
	remote, accountID string,
	adminSite bool,
) (attempts int, autherr []auth.AuthErrors, err error) {
	if adminSite {
		return 0, nil, nil
	}

	err = srv.Repo.AuthAttemptInsert(ctx, &model.AuthAttempt{
		Remote: remote,
	})
	if err != nil {
		return
	}

	attempts, err = srv.Repo.AuthAttemptCountByRemote(ctx, remote)
	if err != nil {
		return
	}

	if attempts > srv.Config.AuthThreshold {
		_, err = srv.BlockUser(ctx, &auth.BlockUserRequest{
			Block: true,
			Id:    accountID,
		})
		if err != nil {
			return
		}

		err = srv.Repo.AuthAttemptRemoveByRemote(ctx, remote)
		if err != nil {
			return
		}

		autherr = []auth.AuthErrors{
			auth.AuthErrors_AUTH_ERROR_ACCOUNT_IS_BLOCKED,
		}
	}

	return
}

func (srv *serverImpl) postLogin(
	ctx context.Context,
	tosAccepted uint64,
	remote, accountID, phone string,
	adminSite bool,
	token *storage.SessionAuth,
) (autherr []auth.AuthErrors, err error) {
	introspect, err := srv.Storage.SessionIntrospect(ctx, token.AccessToken)
	if err != nil {
		log.WithError(err).Error()
		return nil, errInactiveToken
	}

	if !introspect.Active {
		return nil, errInactiveToken
	}

	var tosUpdate, adminRole bool

	for _, r := range introspect.RealmAccess.Roles {
		if r == "admin" {
			adminRole = true
		}

		if r == "tosupdate" {
			tosUpdate = true
		}
	}

	if (adminRole && !adminSite) || (!adminRole && adminSite) {
		log.Debug("user/admin login error")
		return nil, grpcerr.New(grpcerr.NoAuth, "user/admin login error")
	}

	if tosUpdate && tosAccepted < 2 {
		return []auth.AuthErrors{
			auth.AuthErrors_AUTH_ERROR_TOS_REVIEW_REQUIRED,
		}, nil
	}

	err = srv.Repo.AuthAttemptRemoveByRemote(ctx, remote)
	if err != nil {
		return nil, err
	}

	if accountID != "" {
		_, err = srv.Connections.Account.UpdateLastLogin(ctx, &account.UpdateLastLoginRequest{Id: accountID})
		if err != nil {
			log.WithError(err).Error("updating last login")
		}
	}

	known, first, err := srv.Repo.RecordUserLogin(ctx, accountID, remote)
	if err != nil {
		return nil, err
	}

	if !known && !first {
		bs, _ := json.Marshal(map[string]interface{}{
			"remote": remote,
		})

		_, _ = srv.Connections.Notification.SendNotification(ctx, &notification.SendNotificationRequest{
			Entity: notification.Entity_ENTITY_USER_SIGNIN,
			Action: notification.Action_ACTION_CREATE,
			Targets: []*notification.Target{
				{
					User: accountID,
					Role: notification.Role_ROLE_AUTHOR,
					Contacts: &notification.Contacts{
						Phone: phone,
					},
				},
			},
			Embeds: []*notification.EventEmbed{
				{
					EmbedType:  messaging.EmbedType_EMBED_TYPE_JSON,
					EmbedValue: string(bs),
				},
			},
		})
	}

	return nil, nil
}

func (srv *serverImpl) Login(ctx context.Context, req *auth.LoginRequest) (*auth.LoginResponse, error) {
	email := strings.ToLower(req.GetEmailValue())

	acc, autherrs, err := srv.accountLoadEmail(ctx, email, req.GetAdminSiteValue())
	if err != nil || autherrs != nil {
		return &auth.LoginResponse{
			Success: false,
			Errors:  autherrs,
		}, err
	}

	attempts, autherrs, err := srv.preLogin(ctx, req.GetRemoteValue(), acc.GetIdValue(), req.GetAdminSiteValue())
	if err != nil || autherrs != nil {
		return &auth.LoginResponse{
			Success:      false,
			Errors:       autherrs,
			AuthAttempts: uint64(attempts),
		}, err
	}

	token, err := srv.Storage.SessionLogin(ctx, email, req.GetPasswordValue())
	if err != nil {
		log.WithError(err).Error("session login")

		return &auth.LoginResponse{
			Success: false,
			Errors: []auth.AuthErrors{
				auth.AuthErrors_AUTH_ERROR_INVALID_PASSWORD,
			},
			AuthAttempts: uint64(attempts),
		}, nil
	}

	autherrs, err = srv.postLogin(
		ctx,
		req.GetTosAcceptedValue(),
		req.GetRemoteValue(),
		acc.GetIdValue(),
		acc.GetContactValue().GetPhonePrimaryValue(),
		req.GetAdminSiteValue(),
		token,
	)
	if err != nil || autherrs != nil {
		return &auth.LoginResponse{
			Success:      false,
			Errors:       autherrs,
			AuthAttempts: uint64(attempts),
		}, err
	}

	return &auth.LoginResponse{
		Success:      true,
		AccessToken:  token.AccessToken,
		RefreshToken: token.RefreshToken,
	}, nil
}

func (srv *serverImpl) OauthLogin(ctx context.Context, req *auth.OauthRequest) (*auth.OauthResponse, error) {
	users, err := srv.Storage.UserGetByEmail(ctx, req.GetEmailValue())
	if err != nil {
		return nil, err
	}

	if len(users) == 0 {
		return nil, errUserNotFound
	}

	userID := users[0].Id

	ids, err := srv.Repo.UserAttrGetByNameValue(ctx, "oauth", req.GetOauthIdValue())
	if err != nil {
		return nil, err
	}

	if len(ids) > 0 {
		userID = ids[0]
	}

	token, err := srv.Storage.SessionImpersonate(ctx, userID)
	if err != nil {
		return nil, err
	}

	autherrs, err := srv.postLogin(
		ctx,
		req.GetTosAcceptedValue(),
		req.GetRemoteValue(),
		userID,
		"",
		false,
		token,
	)
	if err != nil || autherrs != nil {
		return &auth.OauthResponse{
			Success: false,
			Errors:  autherrs,
		}, err
	}

	return &auth.OauthResponse{
		Success:      true,
		AccessToken:  token.AccessToken,
		RefreshToken: token.RefreshToken,
	}, nil
}

func (srv *serverImpl) Refresh(ctx context.Context, req *auth.RefreshRequest) (*auth.RefreshResponse, error) {
	introspect, err := srv.Storage.SessionIntrospect(ctx, req.GetRefreshToken())
	if err != nil {
		log.WithError(err).Error()
		return nil, errInvalidToken
	}

	if !introspect.Active {
		return nil, errInvalidToken
	}

	token, err := srv.Storage.SessionRefresh(ctx, req.GetRefreshToken())
	if err != nil {
		return nil, err
	}

	return &auth.RefreshResponse{
		Success:      true,
		AccessToken:  token.AccessToken,
		RefreshToken: token.RefreshToken,
	}, nil
}

// verifyUserIntent checks user passsword or pin when one of them is specified.
// NOTE: Not only email can be passed as argument, Keycloak's username value should work too.
func (srv *serverImpl) verifyUserIntent(ctx context.Context, password, pin, email, userID string) []auth.AuthErrors {
	switch {
	case password != "":
		token, err := srv.Storage.SessionLogin(ctx, email, password)
		if err == nil {
			err = srv.Storage.SessionRemove(ctx, token.SessionState)
			if err == nil {
				return nil
			}
		}

		return []auth.AuthErrors{
			auth.AuthErrors_AUTH_ERROR_VERIFICATION_INVALID,
		}
	case pin != "":
		user, err := srv.Storage.UserGetByID(ctx, userID)
		if err == nil {
			if attr, ok := user.Attributes["pin"]; ok {
				if attrstr, ok := attr.(string); ok && len(attrstr) > 0 && attrstr == pin {
					return nil
				}
			}
		}

		return []auth.AuthErrors{
			auth.AuthErrors_AUTH_ERROR_VERIFICATION_INVALID,
		}
	}

	return []auth.AuthErrors{
		auth.AuthErrors_AUTH_ERROR_VERIFICATION_REQUIRED,
	}
}

func (srv *serverImpl) Logout(ctx context.Context, req *auth.LogoutRequest) (*auth.LogoutResponse, error) {
	err := srv.Storage.SessionRemove(ctx, req.GetSession())
	if err != nil {
		return nil, err
	}

	return &auth.LogoutResponse{
		Success: true,
	}, nil
}

func (srv *serverImpl) LogoutAll(ctx context.Context, req *auth.LogoutAllRequest) (*auth.LogoutAllResponse, error) {
	token, err := srv.Storage.SessionIntrospect(ctx, req.GetAccessToken())
	if err != nil {
		log.WithError(err).Error()
		return nil, errInactiveToken
	}

	if !token.Active {
		return nil, errInactiveToken
	}

	accerrs := srv.verifyUserIntent(ctx, req.GetPasswordValue(), req.GetPinValue(), token.Email, token.Sub)
	if accerrs != nil {
		return &auth.LogoutAllResponse{
			Success: false,
			Errors:  accerrs,
		}, nil
	}

	err = srv.Storage.SessionRemoveAll(ctx, token.Sub)
	if err != nil {
		return nil, err
	}

	return &auth.LogoutAllResponse{
		Success: true,
	}, nil
}

func (srv *serverImpl) Signup(ctx context.Context, req *auth.SignupRequest) (*auth.SignupResponse, error) {
	create, err := srv.Create(ctx, req)
	if err != nil {
		return nil, err
	}

	if create.GetSuccess() {
		if req.GetAccount().GetContactValue().GetEmailPrimary() != nil {
			_, _ = srv.Connections.Notification.SendNotification(ctx, &notification.SendNotificationRequest{
				Entity: notification.Entity_ENTITY_USER,
				Action: notification.Action_ACTION_CREATE,
				Targets: []*notification.Target{
					{
						User: create.GetAccount().GetIdValue(),
						Role: notification.Role_ROLE_AUTHOR,
					},
				},
			})
		}
	}

	return &auth.SignupResponse{
		Success:      create.GetSuccess(),
		Errors:       create.GetErrors(),
		AccessToken:  create.GetAccessToken(),
		Valid:        create.GetValid(),
		Verification: create.GetVerification(),
		RefreshToken: create.GetRefreshToken(),
	}, nil
}

func (srv *serverImpl) Create(ctx context.Context, req *auth.SignupRequest) (resp *auth.CreateResponse, err error) {
	if req.Account == nil {
		req.Account = &account.Account{}
	}

	if req.Account.Contact == nil {
		req.Account.Contact = &account.Account_ContactValue{ContactValue: &account.AccountContact{}}
	}

	if req.Account.GetContactValue().ContactStatus == nil {
		req.Account.GetContactValue().ContactStatus = &account.AccountContact_ContactStatusValue{
			ContactStatusValue: &account.AccountContactStatus{},
		}
	}

	if req.Account.GetContactValue().EmailPrimary != nil {
		req.Account.GetContactValue().EmailPrimary = &account.AccountContact_EmailPrimaryValue{
			EmailPrimaryValue: strings.ToLower(req.GetAccount().GetContactValue().GetEmailPrimaryValue()),
		}
	}

	if req.Account.GetContactValue().PhonePrimary != nil {
		req.Account.GetContactValue().PhonePrimary = &account.AccountContact_PhonePrimaryValue{
			PhonePrimaryValue: strings.ToLower(req.GetAccount().GetContactValue().GetPhonePrimaryValue()),
		}
	}

	pass := req.GetPasswordValue()
	if pass == "" {
		pass, err = srv.generateDefaultPassword()
		if err != nil {
			log.WithError(err).Error("error while generating new default password")
		}
	}

	if pass != "" {
		req.Password = &auth.SignupRequest_PasswordValue{
			PasswordValue: pass,
		}
	}

	valid, verif, autherrs, err := srv.validateCreateAccount(
		ctx,
		req.GetAccount(),
		req.GetPasswordValue(),
		req.GetOauthIdValue(),
		req.GetOauthTypeValue(),
		req.GetRemoteValue(),
		req.GetPhoneOtpValue(),
		req.GetAdminSiteValue(),
		req.GetAccount().GetAdminValue(),
		req.GetTosAcceptedValue(),
	)
	if err != nil || len(autherrs) > 0 || !valid {
		return &auth.CreateResponse{
			Success:      false,
			Errors:       autherrs,
			Valid:        valid,
			Verification: verif,
		}, err
	}

	primary := strings.ToLower(req.GetAccount().GetContactValue().GetEmailPrimaryValue())

	username := primary
	if len(username) == 0 {
		username = req.GetOauthIdValue()
	}

	var disableableCredentialTypes []string

	var credentials []keycloak.CredentialRepresentation

	if req.GetPasswordValue() != "" {
		disableableCredentialTypes = append(disableableCredentialTypes, "password")
		credentials = append(credentials, keycloak.CredentialRepresentation{
			Type:      "password",
			Value:     req.GetPasswordValue(),
			Temporary: false,
		})
	}

	userinfo := &storage.User{
		Username:  username,
		Enabled:   true,
		Email:     req.GetAccount().GetContactValue().GetEmailPrimaryValue(),
		FirstName: req.GetAccount().GetDetailsValue().GetFirstNameValue(),
		LastName:  req.GetAccount().GetDetailsValue().GetLastNameValue(),
		Attributes: map[string]interface{}{
			"oauth_type":        req.GetOauthTypeValue(),
			"oauth":             req.GetOauthIdValue(),
			"email":             primary,
			"pin":               req.GetPinValue(),
			"transient_session": req.GetTransientSessionValue(),
			"phoneNumber":       normalizePhoneNumber(req.GetAccount().GetContactValue().GetPhonePrimaryValue()),
		},
		DisableableCredentialTypes: disableableCredentialTypes,
		Access: map[string]interface{}{
			"manageGroupMembership": true,
			"view":                  true,
			"mapRoles":              true,
			"impersonate":           true,
			"manage":                true,
		},
		Credentials: credentials,
	}

	err = srv.Storage.UserCreate(ctx, userinfo)
	if err != nil {
		return &auth.CreateResponse{
			Success:      false,
			Valid:        valid,
			Verification: verif,
		}, err
	}

	token, err := srv.Storage.SessionImpersonate(ctx, username)
	if err != nil {
		return nil, err
	}

	introspect, err := srv.Storage.SessionIntrospect(ctx, token.AccessToken)
	if err != nil {
		return nil, err
	}

	if req.GetAdminSiteValue() && req.GetAccount().GetAdminValue() {
		err = srv.setAdminRole(ctx, introspect)
		if err != nil {
			return &auth.CreateResponse{
				Success:      false,
				Valid:        valid,
				Verification: verif,
			}, err
		}
	}

	if req.Account == nil {
		req.Account = &account.Account{}
	}

	req.GetAccount().Id = &account.Account_IdValue{IdValue: introspect.Sub}
	req.GetAccount().TosVersion = &account.Account_TosVersionValue{TosVersionValue: int64(req.GetTosAcceptedValue())}

	creationType := account.AccountCreationType_ACCOUNT_CREATION_TYPE_EMAIL
	if req.GetOauthTypeValue() == "google" {
		creationType = account.AccountCreationType_ACCOUNT_CREATION_TYPE_GOOGLE
	}

	req.GetAccount().CreationType = &account.Account_CreationTypeValue{CreationTypeValue: creationType}

	acc, err := srv.Connections.Account.Create(ctx, &account.CreateRequest{
		Account: req.GetAccount(),
	})
	if acc == nil {
		acc = &account.CreateResponse{}
	}

	if err != nil || !acc.GetValid() {
		return &auth.CreateResponse{
			Verification: acc.GetVerification(),
		}, err
	}

	if req.GetAdminSiteValue() {
		token.AccessToken, token.RefreshToken = "", ""
	}

	return &auth.CreateResponse{
		Account:      acc.Account,
		Success:      acc.Valid,
		Errors:       autherrs,
		Valid:        acc.Valid,
		Verification: acc.Verification,
		AccessToken:  token.AccessToken,
		RefreshToken: token.RefreshToken,
	}, nil
}

func (srv *serverImpl) setAdminRole(ctx context.Context, introspect *storage.SessionIntrospection) error {
	role, err := srv.Storage.RoleGetByCode(ctx, "admin")
	if err != nil {
		return err
	}

	err = srv.Storage.UserUpdateRoles(ctx, introspect.Sub, []*storage.Role{role})
	if err != nil {
		return err
	}

	return nil
}

func (srv *serverImpl) updateUserDetails(
	ctx context.Context,
	user *storage.User,
	acc *account.Account,
	newPassword, email, phone string,
) (err error) {
	var update, phoneChange, emailChange bool

	if user.Email == "" && user.Username != "" {
		user.Email = user.Username
		user.Attributes["email"] = user.Username
		update = true
	}

	if acc.GetContactValue().GetEmailPrimaryValue() != "" {
		primary := strings.ToLower(acc.GetContactValue().GetEmailPrimaryValue())
		user.Username = primary
		user.Email = primary
		user.Attributes["email"] = primary
		update = true
		emailChange = true
	}

	if acc.GetContactValue().GetPhonePrimaryValue() != "" {
		user.Attributes["phoneNumber"] = normalizePhoneNumber(acc.GetContactValue().GetPhonePrimaryValue())
		update = true
		phoneChange = true
	}

	if acc.GetDetailsValue().GetFirstNameValue() != "" {
		user.FirstName = acc.GetDetailsValue().GetFirstNameValue()
		update = true
	}

	if acc.GetDetailsValue().GetLastNameValue() != "" {
		user.LastName = acc.GetDetailsValue().GetLastNameValue()
		update = true
	}

	if newPassword != "" {
		user.Credentials = []keycloak.CredentialRepresentation{
			{
				Type:      "password",
				Value:     newPassword,
				Temporary: false,
			},
		}
		update = true
	}

	if update {
		err = srv.Storage.UserUpdate(ctx, user)
		if err != nil {
			return err
		}
	}

	if emailChange {
		err = srv.notifyPrimaryEmailChanged(
			ctx,
			acc.GetContactValue().GetEmailPrimaryValue(),
			email,
		)
		if err != nil {
			log.WithError(err).Error()
		}
	}

	if phoneChange {
		err = srv.notifyPrimaryPhoneChanged(
			ctx,
			normalizePhoneNumber(acc.GetContactValue().GetPhonePrimaryValue()),
			phone,
			email,
		)
		if err != nil {
			log.WithError(err).Error()
		}
	}

	return nil
}

func (srv *serverImpl) Update(ctx context.Context, req *auth.UpdateRequest) (*auth.UpdateResponse, error) {
	introspect, err := srv.Storage.SessionIntrospect(ctx, req.GetAccessToken())
	if err != nil {
		return nil, err
	}

	if !introspect.Active {
		return nil, errInactiveToken
	}

	user, err := srv.Storage.UserGetByID(ctx, introspect.Sub)
	if err != nil {
		return nil, err
	}

	if req.Account == nil {
		req.Account = &account.Account{}
	}

	if req.Account.Contact == nil {
		req.Account.Contact = &account.Account_ContactValue{ContactValue: &account.AccountContact{}}
	}

	if req.Account.GetContactValue().ContactStatus == nil {
		req.Account.GetContactValue().ContactStatus = &account.AccountContact_ContactStatusValue{
			ContactStatusValue: &account.AccountContactStatus{},
		}
	}

	if req.Account.GetContactValue().EmailPrimary != nil {
		req.Account.GetContactValue().EmailPrimary = &account.AccountContact_EmailPrimaryValue{
			EmailPrimaryValue: strings.ToLower(req.GetAccount().GetContactValue().GetEmailPrimaryValue()),
		}
	}

	current, autherrs, err := srv.accountLoadID(ctx, introspect.Sub, false)
	if err != nil || len(autherrs) > 0 {
		return &auth.UpdateResponse{
			Success: true,
			Errors:  autherrs,
		}, err
	}

	valid, verif, autherrs, err := srv.validateUpdateAccount(
		ctx,
		introspect,
		current,
		req.GetAccount(),
		req.GetPasswordValue(),
		req.GetNewPasswordValue(),
		req.GetPinValue(),
		req.GetPhoneOtpValue(),
		req.GetEmailOtpValue(),
	)
	if err != nil || len(autherrs) > 0 || !valid {
		return &auth.UpdateResponse{
			Success:      false,
			Errors:       autherrs,
			Valid:        valid,
			Verification: verif,
		}, err
	}

	if req.GetAccount() != nil {
		var upd *account.UpdateResponse

		upd, err = srv.Connections.Account.Update(ctx, &account.UpdateRequest{
			Account: req.GetAccount(),
		})
		if upd == nil {
			upd = &account.UpdateResponse{}
		}

		if err != nil || !upd.GetValid() {
			return &auth.UpdateResponse{
				Verification: upd.GetVerification(),
			}, err
		}
	}

	err = srv.updateUserDetails(
		ctx,
		user,
		req.GetAccount(),
		req.GetNewPasswordValue(),
		current.GetContactValue().GetEmailPrimaryValue(),
		normalizePhoneNumber(current.GetContactValue().GetPhonePrimaryValue()),
	)
	if err != nil {
		return nil, err
	}

	token, err := srv.Storage.SessionImpersonate(ctx, user.Id)
	if err != nil {
		return nil, err
	}

	return &auth.UpdateResponse{
		Success:      true,
		Valid:        valid,
		Verification: verif,
		AccessToken:  token.AccessToken,
		RefreshToken: token.RefreshToken,
	}, nil
}

func (srv *serverImpl) ForgotPassword(
	ctx context.Context,
	req *auth.ForgotPasswordRequest,
) (*auth.ForgotPasswordResponse, error) {
	email := strings.ToLower(req.GetEmailValue())

	users, err := srv.Storage.UserGetByEmail(ctx, email)
	if err != nil {
		return nil, err
	}

	if len(users) == 0 {
		return nil, errUserNotFound
	}

	accresp, err := srv.Connections.Account.Get(ctx, &account.GetRequest{Id: users[0].Id, Admin: false})
	if errors.Is(err, grpcerr.NotFound) {
		return nil, errUserNotFound
	}

	if err != nil {
		return nil, err
	}

	acc := accresp.GetAccount()

	if acc.Contact == nil {
		acc.Contact = &account.Account_ContactValue{ContactValue: &account.AccountContact{}}
	}

	if acc.GetContactValue().ContactStatus == nil {
		acc.GetContactValue().ContactStatus = &account.AccountContact_ContactStatusValue{
			ContactStatusValue: &account.AccountContactStatus{},
		}
	}

	hash := sha256.Sum256([]byte(
		strings.Join([]string{
			req.GetNewPasswordValue(),
			req.GetEmailOtpValue(),
			req.GetPhoneOtpValue(),
			req.GetCaptchaValue(),
			req.GetRemoteValue(),
		}, ":"),
	))

	data := hex.EncodeToString(hash[:])

	t, err := srv.Repo.ResetAttemptGetByRemoteData(ctx, req.GetRemoteValue(), data)
	if err != nil {
		return nil, err
	}

	if t != nil && time.Since(t.Time) < time.Minute {
		return nil, errResetTooSoon
	}

	err = srv.Repo.ResetAttemptInsert(ctx, &model.ResetAttempt{
		Remote: req.GetRemoteValue(),
		Data:   data,
	})
	if err != nil {
		return nil, err
	}

	autherrs, err := srv.validatePasswordReset(
		ctx,
		acc,
		req.GetNewPasswordValue(),
		req.GetEmailOtpValue(),
	)
	if err != nil || len(autherrs) > 0 {
		return &auth.ForgotPasswordResponse{
			Success: false,
			Errors:  autherrs,
		}, err
	}

	err = srv.Storage.UserPasswordReset(ctx, users[0].Id, req.GetNewPasswordValue())
	if err != nil {
		return nil, err
	}

	_, err = srv.BlockUser(ctx, &auth.BlockUserRequest{
		Id:    users[0].Id,
		Block: false,
	})
	if err != nil {
		return nil, err
	}

	err = srv.Repo.ResetAttemptRemoveByRemote(ctx, req.GetRemoteValue())
	if err != nil {
		return nil, err
	}

	err = srv.Storage.SessionRemoveAll(ctx, users[0].Id)
	if err != nil {
		return nil, err
	}

	token, err := srv.Storage.SessionImpersonate(ctx, users[0].Id)
	if err != nil {
		return nil, err
	}

	return &auth.ForgotPasswordResponse{
		Success:      true,
		AccessToken:  token.AccessToken,
		RefreshToken: token.RefreshToken,
	}, nil
}

func (srv *serverImpl) UpdatePin(ctx context.Context, req *auth.UpdatePinRequest) (*auth.UpdatePinResponse, error) {
	pin := strings.Map(func(r rune) rune {
		if r >= '0' || r <= '9' {
			return r
		}

		return -1
	}, req.GetPinValue())

	if len(pin) != 5 {
		return &auth.UpdatePinResponse{
			Success: false,
			Errors: []auth.AuthErrors{
				auth.AuthErrors_AUTH_ERROR_PIN_INVALID,
			},
		}, nil
	}

	introspect, err := srv.Storage.SessionIntrospect(ctx, req.GetAccessToken())
	if err != nil {
		return nil, err
	}

	if !introspect.Active {
		return nil, errInactiveToken
	}

	acc, _, err := srv.accountLoadID(ctx, introspect.Sub, false)
	if err != nil {
		return nil, err
	}

	if acc.Contact == nil {
		acc.Contact = &account.Account_ContactValue{ContactValue: &account.AccountContact{}}
	}

	if acc.GetContactValue().ContactStatus == nil {
		acc.GetContactValue().ContactStatus = &account.AccountContact_ContactStatusValue{
			ContactStatusValue: &account.AccountContactStatus{},
		}
	}

	autherrs, err := srv.validateOTP(
		ctx,
		introspect.Sub,
		normalizePhoneNumber(acc.GetContactValue().GetPhonePrimaryValue()),
		req.GetPhoneOtpValue(),
		invitation.ContactValidationType_CONTACT_VALIDATION_PHONE,
		acc.GetContactValue().GetContactStatusValue(),
		nil,
		false,
	)
	if err != nil || len(autherrs) > 0 {
		return &auth.UpdatePinResponse{
			Success: false,
			Errors:  autherrs,
		}, err
	}

	user, err := srv.Storage.UserGetByID(ctx, introspect.Sub)
	if err != nil {
		return nil, err
	}

	user.Attributes["pin"] = pin

	err = srv.Storage.UserUpdate(ctx, user)
	if err != nil {
		return nil, err
	}

	token, err := srv.Storage.SessionImpersonate(ctx, user.Id)
	if err != nil {
		return nil, err
	}

	return &auth.UpdatePinResponse{
		Success:      true,
		AccessToken:  token.AccessToken,
		RefreshToken: token.RefreshToken,
	}, nil
}

func (srv *serverImpl) VerifyPhone(
	ctx context.Context,
	req *auth.VerifyPhoneRequest,
) (*auth.VerifyPhoneResponse, error) {
	introspect, err := srv.Storage.SessionIntrospect(ctx, req.GetAccessToken())
	if err != nil {
		return nil, err
	}

	if !introspect.Active {
		return nil, errInactiveToken
	}

	acc, _, err := srv.accountLoadID(ctx, introspect.Sub, false)
	if err != nil {
		return nil, err
	}

	user, err := srv.Storage.UserGetByID(ctx, introspect.Sub)
	if err != nil {
		return nil, err
	}

	if req.PhoneNumber == nil {
		req.PhoneNumber = &auth.VerifyPhoneRequest_PhoneNumberValue{
			PhoneNumberValue: acc.GetContactValue().GetPhonePrimaryValue(),
		}
	}

	resp, err := srv.Connections.Invitation.ContactValidationVerify(ctx, &invitation.ContactValidationVerifyRequest{
		Type:   invitation.ContactValidationType_CONTACT_VALIDATION_PHONE,
		UserId: introspect.Sub,
		Key:    normalizePhoneNumber(req.GetPhoneNumberValue()),
		Otp:    req.GetPhoneOtpValue(),
	})
	if err != nil {
		return nil, err
	}

	var autherrs []auth.AuthErrors

	switch resp.GetStatus() {
	case invitation.ContactValidationStatus_CONTACT_VALIDATION_STATUS_OK:
	case invitation.ContactValidationStatus_CONTACT_VALIDATION_STATUS_EXPIRED:
		autherrs = append(autherrs, auth.AuthErrors_AUTH_ERROR_PHONE_OTP_EXPIRED)
	default:
		autherrs = append(autherrs, auth.AuthErrors_AUTH_ERROR_PHONE_OTP_INVALID)
	}

	if len(autherrs) > 0 {
		return &auth.VerifyPhoneResponse{
			Success: false,
			Errors:  autherrs,
		}, nil
	}

	user.Attributes["phoneNumber"] = normalizePhoneNumber(req.GetPhoneNumberValue())

	err = srv.Storage.UserUpdate(ctx, user)
	if err != nil {
		return nil, err
	}

	_, err = srv.Connections.Account.Update(ctx, &account.UpdateRequest{
		Account: &account.Account{
			Id: &account.Account_IdValue{
				IdValue: acc.GetIdValue(),
			},
			Contact: &account.Account_ContactValue{
				ContactValue: &account.AccountContact{
					PhonePrimary: &account.AccountContact_PhonePrimaryValue{
						PhonePrimaryValue: normalizePhoneNumber(req.GetPhoneNumberValue()),
					},
					ContactStatus: &account.AccountContact_ContactStatusValue{
						ContactStatusValue: &account.AccountContactStatus{
							PhoneVerified: &account.AccountContactStatus_PhoneVerifiedValue{
								PhoneVerifiedValue: true,
							},
						},
					},
				},
			},
		},
	})
	if err != nil {
		return nil, err
	}

	return &auth.VerifyPhoneResponse{
		Success: true,
	}, nil
}

func (srv *serverImpl) VerifyEmail(
	ctx context.Context,
	req *auth.VerifyEmailRequest,
) (*auth.VerifyEmailResponse, error) {
	introspect, err := srv.Storage.SessionIntrospect(ctx, req.GetAccessToken())
	if err != nil {
		return nil, err
	}

	if !introspect.Active {
		return nil, errInactiveToken
	}

	acc, _, err := srv.accountLoadID(ctx, introspect.Sub, false)
	if err != nil {
		return nil, err
	}

	resp, err := srv.Config.Connections.Invitation.ContactValidationVerify(ctx, &invitation.ContactValidationVerifyRequest{
		Type:   invitation.ContactValidationType_CONTACT_VALIDATION_EMAIL,
		UserId: introspect.Sub,
		Key:    acc.GetContactValue().GetEmailPrimaryValue(),
		Otp:    req.GetEmailOtpValue(),
	})
	if err != nil {
		return nil, err
	}

	var autherrs []auth.AuthErrors

	switch resp.GetStatus() {
	case invitation.ContactValidationStatus_CONTACT_VALIDATION_STATUS_OK:
	case invitation.ContactValidationStatus_CONTACT_VALIDATION_STATUS_EXPIRED:
		autherrs = append(autherrs, auth.AuthErrors_AUTH_ERROR_EMAIL_OTP_EXPIRED)
	default:
		autherrs = append(autherrs, auth.AuthErrors_AUTH_ERROR_EMAIL_OTP_INVALID)
	}

	if len(autherrs) > 0 {
		return &auth.VerifyEmailResponse{
			Success: false,
			Errors:  autherrs,
		}, nil
	}

	_, _ = srv.Connections.Account.Update(ctx, &account.UpdateRequest{
		Account: &account.Account{
			Id: &account.Account_IdValue{
				IdValue: acc.GetIdValue(),
			},
			Contact: &account.Account_ContactValue{
				ContactValue: &account.AccountContact{
					ContactStatus: &account.AccountContact_ContactStatusValue{
						ContactStatusValue: &account.AccountContactStatus{
							EmailVerified: &account.AccountContactStatus_EmailVerifiedValue{
								EmailVerifiedValue: true,
							},
						},
					},
				},
			},
		},
	})

	return &auth.VerifyEmailResponse{
		Success: true,
	}, nil
}

func (srv *serverImpl) BlockUser(ctx context.Context, req *auth.BlockUserRequest) (*auth.BlockUserResponse, error) {
	user, err := srv.Storage.UserGetByID(ctx, req.GetId())
	if err != nil {
		return nil, err
	}

	user.Enabled = !req.GetBlock()

	err = srv.Storage.UserUpdate(ctx, user)
	if err != nil {
		return nil, err
	}

	resp, err := srv.Connections.Account.Block(ctx, &account.BlockRequest{
		Id:    req.GetId(),
		Block: req.GetBlock(),
	})
	if err != nil {
		return nil, err
	}

	return &auth.BlockUserResponse{
		Success: resp.GetSuccess(),
	}, nil
}

func (srv *serverImpl) CheckPassword(
	ctx context.Context,
	req *auth.CheckPasswordRequest,
) (*auth.CheckPasswordResponse, error) {
	token, err := srv.Storage.SessionLogin(ctx, req.GetEmail(), req.GetPassword())
	if err != nil {
		return nil, err
	}

	err = srv.Storage.SessionRemove(ctx, token.SessionState)
	if err != nil {
		return nil, err
	}

	return &auth.CheckPasswordResponse{
		Success: true,
	}, nil
}

func (srv *serverImpl) ResetAllAuthAttempts(
	ctx context.Context,
	req *auth.ResetAllAuthAttemptsRequest,
) (resp *auth.ResetAllAuthAttemptsResponse, err error) {
	if req.GetIpValue() != "" {
		err = srv.Repo.ResetAttemptRemoveByRemote(ctx, req.GetIpValue())
	} else {
		err = srv.Repo.ResetAttemptRemoveAll(ctx)
	}

	if err != nil {
		return nil, err
	}

	return &auth.ResetAllAuthAttemptsResponse{
		Success: true,
	}, nil
}

func (srv *serverImpl) ServiceUpdatePassword(
	ctx context.Context,
	req *auth.ServiceUpdatePasswordRequest,
) (*auth.ServiceUpdatePasswordResponse, error) {
	autherrs := srv.validatePassword(req.GetNewPassword())
	if len(autherrs) > 0 {
		return &auth.ServiceUpdatePasswordResponse{
			Success: false,
			Errors:  autherrs,
		}, nil
	}

	acc, err := srv.Connections.Account.Get(ctx, &account.GetRequest{Id: req.GetUserId()})
	if err != nil {
		return nil, err
	}

	err = srv.Storage.UserPasswordReset(ctx, req.GetUserId(), req.GetNewPassword())
	if err != nil {
		return nil, err
	}

	if acc.GetAccount().GetBlockValue() {
		_, err = srv.BlockUser(ctx, &auth.BlockUserRequest{
			Id:    req.GetUserId(),
			Block: false,
		})
		if err != nil {
			return nil, err
		}
	}

	return &auth.ServiceUpdatePasswordResponse{
		Success: true,
	}, nil
}

func (srv *serverImpl) generateDefaultPassword() (string, error) {
	const (
		chars          = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
		passwordLength = 16
	)

	bytes := make([]byte, passwordLength)
	if _, err := rand.Read(bytes); err != nil {
		return "", err
	}

	for i, b := range bytes {
		bytes[i] = chars[b%byte(len(chars))]
	}

	return string(bytes), nil
}
